//
// TinyPTC by Gaffer
// www.gaffer.org/tinyptc
//
#include "convert.h"
#ifdef __PTC_MMX__
#include "mmx.h"
#endif
#include <memory.h>

void ptc_convert_32_to_32_rgb888(void *src,void *dst,int pixels)
{
    #ifdef __PTC_MMX__

        mmx_memcpy(dst,src,pixels*4);

    #else

        #ifdef __PTC_MEMCPY__

            memcpy(dst,src,pixels*4);

        #else

            int32 *p = (int32*) src;
            int32 *q = (int32*) dst;
            while (pixels--)
            {
                *(q++) = *(p++);
            }

        #endif


    #endif
}

void ptc_convert_32_to_32_bgr888(void *src,void *dst,int pixels)
{
    int32 *p = (int32*) src;
    int32 *q = (int32*) dst;
    while (pixels--)
    {
        int32 r = (int32) ( (*p & 0x00FF0000) >> 16 );
        int32 g = (int32) ( (*p & 0x0000FF00) );
        int32 b = (int32) ( (*p & 0x000000FF) << 16 );
        *(q++) = r | g | b;
    }
}

void ptc_convert_32_to_24_rgb888(void *src,void *dst,int pixels)
{
    #ifdef __PTC_MMX__

        mmx_convert_32_to_24_rgb888(dst,src,pixels);

    #else

        char8 *p = (char8*) src;
        char8 *q = (char8*) dst;
        while (pixels--)
        {
            #ifdef __PTC_LITTLE_ENDIAN__
            *q = *p;
            *(q+1) = *(p+1);
            *(q+2) = *(p+2);
            #else
            *(q+2) = *p;
            *(q+1) = *(p+1);
            *q = *(p+2);
            #endif
            p+=4;
            q+=3;
        }

    #endif
}

void ptc_convert_32_to_24_bgr888(void *src,void *dst,int pixels)
{
    char8 *p = (char8*) src;
    char8 *q = (char8*) dst;
    while (pixels--)
    {
        #ifdef __PTC_LITTLE_ENDIAN__
        *(q+2) = *p;
        *(q+1) = *(p+1);
        *q = *(p+2);
        #else
        *q = *p;
        *(q+1) = *(p+1);
        *(q+2) = *(p+2);
        #endif
        p+=4;
        q+=3;
    }
}

void ptc_convert_32_to_16_rgb565(void *src,void *dst,int pixels)
{
    #ifdef __PTC_MMX__
        mmx_convert_32_to_16_rgb565(dst,src,pixels);

    #else


        int32 *p = (int32*) src;
        short16 *q = (short16*) dst;
        while (pixels--)
        {
            short16 r = (short16) ( (*p & 0x00F80000) >> 8 );
            short16 g = (short16) ( (*p & 0x0000FC00) >> 5 );
            short16 b = (short16) ( (*p & 0x000000F8) >> 3 );
            *q = r | g | b;
            p++;
            q++;
        }

    #endif
}

void ptc_convert_32_to_16_bgr565(void *src,void *dst,int pixels)
{
    #ifdef __PTC_MMX__

        mmx_convert_32_to_16_bgr565(dst,src,pixels);

    #else

        int32 *p = (int32*) src;
        short16 *q = (short16*) dst;
        while (pixels--)
        {
            short16 r = (short16) ( (*p & 0x00F80000) >> 19 );
            short16 g = (short16) ( (*p & 0x0000FC00) >> 5  );
            short16 b = (short16) ( (*p & 0x000000F8) << 8  );
            *q = r | g | b;
            p++;
            q++;
        }

    #endif
}

void ptc_convert_32_to_16_rgb555(void *src,void *dst,int pixels)
{
    #ifdef __PTC_MMX__

        mmx_convert_32_to_16_rgb555(dst,src,pixels);

    #else

        int32 *p = (int32*) src;
        short16 *q = (short16*) dst;
        while (pixels--)
        {
            short16 r = (short16) ( (*p & 0x00F80000) >> 9 );
            short16 g = (short16) ( (*p & 0x0000F800) >> 6 );
            short16 b = (short16) ( (*p & 0x000000F8) >> 3 );
            *q = r | g | b;
            p++;
            q++;
        }

    #endif
}

void ptc_convert_32_to_16_bgr555(void *src,void *dst,int pixels)
{
    #ifdef __PTC_MMX__

        mmx_convert_32_to_16_rgb555(dst,src,pixels);

    #else

        int32 *p = (int32*) src;
        short16 *q = (short16*) dst;
        while (pixels--)
        {
            short16 r = (short16) ( (*p & 0x00F80000) >> 20 );
            short16 g = (short16) ( (*p & 0x0000F800) >> 6  );
            short16 b = (short16) ( (*p & 0x000000F8) << 8  );
            *q = r | g | b;
            p++;
            q++;
        }

    #endif
}

PTC_CONVERTER ptc_request_converter(color_space cs)
{
    // 32bit RGB888 -> 32bit RGB888
    #ifdef __PTC_CONVERTER_32_TO_32_RGB888
    if (cs == B_RGB32) return &ptc_convert_32_to_32_rgb888;
    #endif

    // 32bit RGB888 -> 32bit BGR888
    #ifdef __PTC_CONVERTER_32_TO_32_BGR888
    if (cs == B_RGB32_BIG) return &ptc_convert_32_to_32_bgr888;
    #endif

    // 32bit RGB888 -> 24bit RGB888
    #ifdef __PTC_CONVERTER_32_TO_24_RGB888
    if (cs == B_RGB24) return &ptc_convert_32_to_24_rgb888;
    #endif

    // 32bit RGB888 -> 24bit BGR888
    #ifdef __PTC_CONVERTER_32_TO_24_BGR888
    if (cs == B_RGB24_BIG) return &ptc_convert_32_to_24_bgr888;
    #endif
    
    // 32bit RGB888 -> 16bit RGB565
    #ifdef __PTC_CONVERTER_32_TO_16_RGB565
    if (cs == B_RGB16) return &ptc_convert_32_to_16_rgb565;
    #endif

    // 32bit RGB888 -> 16bit BGR565
    #ifdef __PTC_CONVERTER_32_TO_16_BGR565
    if (cs == B_RGB16_BIG) return &ptc_convert_32_to_16_bgr565;
    #endif

    // 32bit RGB888 -> 16bit RGB555
    #ifdef __PTC_CONVERTER_32_TO_16_RGB555
    if (cs == B_RGB15) return &ptc_convert_32_to_16_rgb555;
    #endif

    // 32bit RGB888 -> 16bit BGR555
    #ifdef __PTC_CONVERTER_32_TO_16_BGR555
    if (cs == B_RGB15_BIG) return &ptc_convert_32_to_16_bgr555;
    #endif

    // failure
    return 0;
}
