/* BeOS version 0.1 of TinyPTC by Johnny BARRAY (jbarray@worldonline.fr)
** The tinyPTC API C fonctions are calling the method of the TinyPTCBeOS object
**
*/
#include "tinyptc.h"
#include "tinyptcBeOS.h" 
#include "convert.h"

#include <DirectWindow.h>
#include <Screen.h>
#include <View.h>

#include <stdlib.h>
#include <string.h>

/************************************************************************************
** The C functions of the tinyptc.h API
** You'll find the 3 one, ptc_open, ptc_close and ptc_update
************************************************************************************/
int ptc_open(char *title,int width,int height)
{	/* Create the TinyPTCBeOS application,
	** See the code in the constructor
	** NOTE : in order to let "as is" the standard use of TinyPTC, I don't call
	**        the be_app->run() method because it stop the current thread untill the app is out.
	**        Of course, there are some problems with this method (be_app doesn't received any message)
	**        but tinyPTC was done like this so I hope the positive thing is that many coders using
	**        tinyPTC will port their work to BeOS :)
	*/
	new TinyPTCBeOS(title, width, height);
	return (1);
}

void ptc_close()
{	/* Delete the TinyPTCBeOS application,
	** See the code in the destructor
	** NOTE : in the current implementation (a infinit loop in the main function)
	**        this function is never called but BeOS will call the destructor...
	*/
	delete ((TinyPTCBeOS*)be_app);
}

int ptc_update(void *buffer)
{	/* Be logic, so call a TinyPTCBeos method...
	*/
	((TinyPTCBeOS*)be_app)->Update(buffer);
	return (1);
}

/************************************************************************************
** Here is the constructor of the TinyPTCBeOS object
** Input parameters are :
**   the title of the window in windowed mode,
**   the width of the demo screen,
**   the height of the demo screen.
** Output parameters are :
**   none
** Object :
**   The constructor of the application create a new TinyPTCWindow object
************************************************************************************/
TinyPTCBeOS::TinyPTCBeOS(const char *title,int width,int height)
 : BApplication("application/x-vnd.Be-TinyPTCSample")
{ 
	pDemoScreen = new TinyPTCWindow(title, width, height);
}

/************************************************************************************
** Here is the destructor of the TinyPTCBeOS object
** Input parameters are :
**   none
** Output parameters are :
**   none
** Object :
**   Dispose ressources
************************************************************************************/
TinyPTCBeOS::~TinyPTCBeOS()
{
	delete pDemoScreen;
}

/************************************************************************************
** Here is the update method of the TinyPTCBeOS object
** Input parameters are :
**   A pointer to the coder's video buffer
** Output parameters are :
**   none
** Object :
**   Ask the TinyPTCWindow object to update the screen
************************************************************************************/
void TinyPTCBeOS::Update(void *buffer)
{
	((TinyPTCWindow*)pDemoScreen)->Update(buffer);
}

/************************************************************************************
** Here is the constructor of the TinyPTCWindow object
** Input parameters are :
**   the title of the window in windowed mode,
**   the width of the demo screen,
**   the height of the demo screen.
** Output parameters are :
**   none
** Object :
**   Create the BDirectWindow, center the window, check for the window mode support,
**   performs initialization.
************************************************************************************/
TinyPTCWindow::TinyPTCWindow(const char *title,int width,int height)
 : BDirectWindow(BRect(0, 0, width - 1, height - 1), title, B_TITLED_WINDOW, B_NOT_RESIZABLE | B_NOT_ZOOMABLE)
{
	BRect frameScreen;
	BScreen* theScreen;
	BView* pView;

	// Create a bounded view to help erasing the screen in fullscreen mode
	pView = new BView(Bounds(), "", B_FOLLOW_ALL, B_WILL_DRAW);
	pView->SetViewColor(0, 0, 0);
	AddChild(pView);

	// Get screen information to center the window in it
	// TODO: ensure that the screen is large enough to handle the window
	theScreen = new BScreen(this);
	frameScreen = theScreen->Frame();
	if (theScreen->IsValid() == true)
		MoveTo((1 + frameScreen.IntegerWidth() - width) / 2,
		       (1 + frameScreen.IntegerHeight() - height) / 2);

	// Go to fullscreen if the videocard doesn't support directwindow in window mode
	// and save the result to allow/prevent the <ALT>-f shortcut
	if (!BDirectWindow::SupportsWindowMode())
	{	SetFullScreen(true);
		bWindowModeAvailable = false;
		bInFullScreenMode = true;
	}
	else
	{	bWindowModeAvailable = true;
		bInFullScreenMode = false;
	}

	// Add the <ALT>-f shortcut to swap between FullScreen and Window Mode (FSWM BMessage)
	AddShortcut('f', B_COMMAND_KEY, new BMessage('FSWM'));	

	// Store technical and BDirectWindow datas (boolean flags, locker, drawing thread, ...)
	demoScreenWidth = width;
	demoScreenHeight = height;
	bNeedUpdate = false;
	pScreenToShow = NULL;
	ptc_current_converter = NULL;
	fClipList = NULL;
	bConnected = false;
	bDestroying = false;
	pLocker = new BLocker();
	pLockerFullScreen = new BLocker();
	DrawingThread = spawn_thread(TinyPTCWindow::DrawDirectWindow, "DrawingThread",
								 B_DISPLAY_PRIORITY, (void *)this);
	resume_thread(DrawingThread);

	// Show the window
	Show();

	// Dispose ressources
	delete theScreen;
}

/************************************************************************************
** Here is the destructor of the TinyPTCWindow object
** Input parameters are :
**   none
** Output parameters are :
**   none
** Object :
**   Dispose resources allocated by the object
**   Force the application to quit
************************************************************************************/
TinyPTCWindow::~TinyPTCWindow()
{	int32 result;

	bDestroying = true;
	Hide();
	Sync();
	wait_for_thread(DrawingThread, &result);

	free(fClipList);
	delete pLockerFullScreen;
	delete pLocker;

	// not the best way to quit the application under BeOS...see the ptc_open comments
	exit(EXIT_SUCCESS);
}

/************************************************************************************
** Here is the message manager of the TinyPTCWindow object
** Input parameters are :
**   The BMessage the window received
** Output parameters are :
**   none
** Object :
**   in current version, just switch between window and fullscreen mode
************************************************************************************/
void TinyPTCWindow::MessageReceived(BMessage* msg)
{
	switch(msg->what)
	{
	case 'FSWM':
		// Only if the videocard support window mode
		if (bWindowModeAvailable)
		{	// Mutual exclusion with the drawing thread
			pLockerFullScreen->Lock();
			if (IsFullScreen())
				SetFullScreen(bInFullScreenMode = false);
			else
				SetFullScreen(bInFullScreenMode = bNeedErase = true);
			pLockerFullScreen->Unlock();
		}
		break;
	default:
		BDirectWindow::MessageReceived(msg);
		break;
	}
}

/************************************************************************************
** Here is the QuitRequested handler of the TinyPTCWindow object
** Input parameters are :
**   none
** Output parameters are :
**   none
** Object :
**   This is the normal code for BeOS. BUT since so run method has not been called
**   the B_QUIT_REQUESTED is never received by the application...
************************************************************************************/
bool TinyPTCWindow::QuitRequested()
{	((TinyPTCBeOS*)be_app)->PostMessage(B_QUIT_REQUESTED);

	// Sometimes, the app crashes when quit in fullscreen mode
	// So go to windowed mode before quitting...
	if (bWindowModeAvailable)
	{
		pLockerFullScreen->Lock();
		if (IsFullScreen())
			SetFullScreen(bInFullScreenMode = false);
		pLockerFullScreen->Unlock();
	}

	return (true);
}

/************************************************************************************
** Here is the DirectConnected method of the TinyPTCBeOS object
** Input parameters are :
**   A pointer to the direct_buffer_info structure
** Output parameters are :
**   none
** Object :
**   A pale copy of the example of Be Inc.
************************************************************************************/
void TinyPTCWindow::DirectConnected(direct_buffer_info *info)
{
	// Do nothing if the window is not ready or if the window is currently destroyed
	if (!bConnected && bDestroying)
		return;

	// Get exclusive access from the drawing thread
	pLocker->Lock();

	// Check the buffer state and store the information for the drawing thread
	switch (info->buffer_state & B_DIRECT_MODE_MASK)
	{
	case B_DIRECT_START :
		bConnected = true;
	case B_DIRECT_MODIFY :
		if (fClipList)
		{	free(fClipList); 
			fClipList = NULL; 
		}
		fNumClipRects = info->clip_list_count;
		fClipList = (clipping_rect *)malloc(fNumClipRects*sizeof(clipping_rect)); 
		if (fClipList)
		{	memcpy(fClipList, info->clip_list,
					fNumClipRects*sizeof(clipping_rect));
			fBits = (uint8 *) info->bits;
			fRowBytes = info->bytes_per_row;
			fPixelBytes = (info->bits_per_pixel) >> 3;
			if (info->pixel_format != fFormat)
			{	fFormat = info->pixel_format;
				ptc_current_converter = ptc_request_converter(fFormat);
			}
			fBounds = info->window_bounds;
		}
		break;
		
	case B_DIRECT_STOP :
		bConnected = false;
		break;

	case B_BUFFER_RESIZED :
		break;

	default :
		break;
	}

	// Free exclusive access
	pLocker->Unlock();
}

/************************************************************************************
** Here is the update method of the TinyPTCWindow object
** Input parameters are :
**   A pointer to the coder's video buffer
** Output parameters are :
**   none
** Object :
**   Backup the pointer to the video buffer of the coder
**   Tell the drawing thread to go on drawing
**   Wait for the drawing thread to complete the draw of the window
**   So the loop coder will be synchronize with the vertical blank
************************************************************************************/
void TinyPTCWindow::Update(void *buffer)
{
	pScreenToShow = (uint32*)buffer;
	bNeedUpdate = true;
	while (bNeedUpdate)
	{	snooze(SNOOZE_TIME);
	};
}

/************************************************************************************
** Here is the Drawing method of the TinyPTCWindow object
** METHOD CALLED FROM THE DRAWING THREAD
** Input parameters are :
**   A pointer to the TinyPTCBeOS object
** Output parameters are :
**   B_OK but when the window is being destroyed...
** Object :
**   This method draw the screen...
************************************************************************************/
int32 TinyPTCWindow::DrawDirectWindow(void *ptr)
{	TinyPTCWindow* PTCWindow = (TinyPTCWindow*)ptr;
	BScreen* theScreen = new BScreen();
	uint32* pPixel;

	// While the window is up
	while (!PTCWindow->bDestroying)
	{
		// Wait for a need of update
		// I'm not sure it is the best way for doing that
		// But the goal of this lib is to be used for demo purpose
		// So the thread will not wait for too long time
		// remember *REAL* DEMO MUST HAVE THE SAME FPS AS THE MONITOR
		// but *real* demo coders know that :)
		while (!PTCWindow->bNeedUpdate)
		{	snooze(SNOOZE_TIME);
		};
		
		// Be sure the window is not changing...
		PTCWindow->pLockerFullScreen->Lock();
		PTCWindow->pLocker->Lock();
		if (PTCWindow->bConnected)
		{
			// There is no converter for the 8 bits per pixel color detph...
			if (PTCWindow->ptc_current_converter != 0)
			{
				int32 y, width, height, adder;
				void *pDst;
				clipping_rect *clip;
				uint32 i;

				// size of a row (in bytes not pixel) of the destination screen
				adder = PTCWindow->fRowBytes;

				// in fullScreenMode, PTCWindow->fNumClipRects always == 1
				for (i = 0; i < PTCWindow->fNumClipRects; i++)
				{
					// In FullScreen mode, the entire demo screen is draw in the middle of the screen
					if (PTCWindow->bInFullScreenMode)
					{
						BRect frameScreen;
						int screenHeight, screenWidth;
						frameScreen = theScreen->Frame();
						screenHeight = frameScreen.IntegerHeight()+1;
						screenWidth = frameScreen.IntegerWidth()+1;
						
						height = PTCWindow->demoScreenHeight;
						width = PTCWindow->demoScreenWidth;

						// Erase the last window image off the screen if need
						if (PTCWindow->bNeedErase)
						{
							y = 0;
							pDst = PTCWindow->fBits;
							while (y < height)
							{
								memset(pDst, 0, PTCWindow->fPixelBytes * width);
								y++;
								pDst = (void *)((int32)pDst + adder);
							}
							PTCWindow->bNeedErase = false;
						}

						// In fullscreen mode, center the drawing
						pDst = PTCWindow->fBits + (PTCWindow->fRowBytes * (screenHeight - height) / 2)
												+ (PTCWindow->fPixelBytes * (screenWidth - width) / 2);
						pPixel = PTCWindow->pScreenToShow;
					}
					else
					// Normal code for windowed display
					{
						// Clip region in screen coordinate
						clip = &(PTCWindow->fClipList[i]);
						width = (clip->right - clip->left)+1;
						height = (clip->bottom - clip->top)+1;
						pDst = PTCWindow->fBits + (clip->top*PTCWindow->fRowBytes)
												+ PTCWindow->fPixelBytes*clip->left;

						// Perform the start adress in the source buffer
						pPixel = PTCWindow->pScreenToShow + (clip->left - PTCWindow->fBounds.left)
									+ ((clip->top - PTCWindow->fBounds.top) * PTCWindow->demoScreenWidth);
					}

					// DirectDraw the rectangle to be update using the current pixel converter
					y = 0;
					while (y < height)
					{	//if (width >= 64)
						(PTCWindow->ptc_current_converter)(pPixel + y * PTCWindow->demoScreenWidth,
													       pDst, width);
						y++;
						pDst = (void *)((int32)pDst + adder);
					}
				}
			}

		}

		// Allow DirectConnect and MessageReceived to do their works
		PTCWindow->pLockerFullScreen->Unlock();
		PTCWindow->pLocker->Unlock();

		// Wait for Vertical Blank to insure smooth animation
		if (theScreen->IsValid() == true)
			theScreen->WaitForRetrace();
		else
			// Or for some driver, wait for some times
			// TODO: ensure 60 FPS synchronisation
			snooze(16000);
		
		// the tinyPTC update function is waiting for this boolean to be false
		PTCWindow->bNeedUpdate = false;
	}

	//  Some tests on my computers with a beta version :
	//	Desktop PIII-500, 320*200, WaitForRetrace() is working 
	//	Mode				PerformAnimation	DrawAnimation	fps
	//---------------------------------------------------------------
	//	32Bits MMX				4279				1158		 75
	//	32Bits MEMCPY			4261				1188		 75
	//  32Bits "C"				4255				2007		 75
	//
	//	16Bits MMX				4271				1126		 75
	//	16Bits "C"				4349				5801		~74
	//
	//	15Bits MMX				4253				 897		 75
	//	15Bits "C"				4290				5668		 75
	
	//	Laptop PII-366, 320*200, WaitForRetrace() isn't working 
	//	Mode				PerformAnimation	DrawAnimation	fps
	//---------------------------------------------------------------
	//	16Bits MMX				6578				5548		 80
	//	16Bits "C"				6803			   15421		 45
	
	return B_OK;		
}
